

if __name__ == '__main__':
	texts = ['1', '2', '3', '4', '5']
	length = len(texts)
	batch_size = 4
	n = length / batch_size

	print([texts[i:i + batch_size] for i in range(0, len(texts), batch_size)])


import torch
from pathlib import Path
from joblib import Parallel, delayed
from typing import Union, List, Optional
from transformers import BertTokenizer, BertConfig, BertModel, AutoModel
from transformers.tokenization_utils import PaddingStrategy, TruncationStrategy, TensorType


class TokenizeVec:

	def __init__(self, pretrained_path: Union[str, Path]):
		self.pretrained = None
		self.tokenizer = BertTokenizer.from_pretrained(pretrained_path)

	def encode_plus(self,
				texts: List[str],
				add_special_tokens: bool = True,
				padding: Union[bool, str, PaddingStrategy] = False,
				truncation: Union[bool, str, TruncationStrategy] = None,
				max_length: Optional[int] = None,
				stride: int = 0,
				is_split_into_words: bool = False,
				pad_to_multiple_of: Optional[int] = None,
				return_tensors: Optional[Union[str, TensorType]] = None,
				return_token_type_ids: Optional[bool] = None,
				return_attention_mask: Optional[bool] = None,
				return_overflowing_tokens: bool = False,
				return_special_tokens_mask: bool = False,
				return_offsets_mapping: bool = False,
				return_length: bool = False,
				verbose: bool = True,
				cls: bool = False,
				**kwargs,
			) -> torch.FloatTensor:
		tokens = self.tokenizer.batch_encode_plus(
												texts,
												add_special_tokens,
												padding,
												truncation,
												max_length,
												stride,
												is_split_into_words,
												pad_to_multiple_of,
												return_tensors,
												return_token_type_ids,
												return_attention_mask,
												return_overflowing_tokens,
												return_special_tokens_mask,
												return_offsets_mapping,
												return_length,
												verbose,
												** kwargs
											)
		self.pretrained.eval()
		with torch.no_grad():
			output = self.pretrained(**tokens, output_hidden_states=True)
		return output.last_hidden_state[:, 0] if cls else output.last_hidden_state[:, 1:]

	def batch_encode_plus(self,
						texts: List[str],
						add_special_tokens: bool = True,
						padding: Union[bool, str, PaddingStrategy] = False,
						truncation: Union[bool, str, TruncationStrategy] = None,
						max_length: Optional[int] = None,
						stride: int = 0,
						is_split_into_words: bool = False,
						pad_to_multiple_of: Optional[int] = None,
						return_tensors: Optional[Union[str, TensorType]] = None,
						return_token_type_ids: Optional[bool] = None,
						return_attention_mask: Optional[bool] = None,
						return_overflowing_tokens: bool = False,
						return_special_tokens_mask: bool = False,
						return_offsets_mapping: bool = False,
						return_length: bool = False,
						verbose: bool = True,
						cls: bool = False,
						batch_size: int = 128,
						**kwargs,
					) -> torch.FloatTensor:
		length = len(texts)
		if length <= batch_size:
			return self.encode_plus(
									texts,
									add_special_tokens,
									padding,
									truncation,
									max_length,
									stride,
									is_split_into_words,
									pad_to_multiple_of,
									return_tensors,
									return_token_type_ids,
									return_attention_mask,
									return_overflowing_tokens,
									return_special_tokens_mask,
									return_offsets_mapping,
									return_length,
									verbose,
									cls,
									** kwargs
								)
		else:
			text_list = self.split_texts(texts, batch_size)
			results = [self.encode_plus(
									text,
									add_special_tokens,
									padding,
									truncation,
									max_length,
									stride,
									is_split_into_words,
									pad_to_multiple_of,
									return_tensors,
									return_token_type_ids,
									return_attention_mask,
									return_overflowing_tokens,
									return_special_tokens_mask,
									return_offsets_mapping,
									return_length,
									verbose,
									cls,
									** kwargs
								) for text in text_list]
			return torch.concat(results, dim=0)

	def parallel_encode_plus(self,
						texts: List[str],
						add_special_tokens: bool = True,
						padding: Union[bool, str, PaddingStrategy] = False,
						truncation: Union[bool, str, TruncationStrategy] = None,
						max_length: Optional[int] = None,
						stride: int = 0,
						is_split_into_words: bool = False,
						pad_to_multiple_of: Optional[int] = None,
						return_tensors: Optional[Union[str, TensorType]] = None,
						return_token_type_ids: Optional[bool] = None,
						return_attention_mask: Optional[bool] = None,
						return_overflowing_tokens: bool = False,
						return_special_tokens_mask: bool = False,
						return_offsets_mapping: bool = False,
						return_length: bool = False,
						verbose: bool = True,
						cls: bool = False,
						batch_size: int = 128,
						n_jobs: int = -1,
						**kwargs,
					) -> torch.FloatTensor:
		length = len(texts)
		if length <= batch_size:
			return self.encode_plus(
									texts,
									add_special_tokens,
									padding,
									truncation,
									max_length,
									stride,
									is_split_into_words,
									pad_to_multiple_of,
									return_tensors,
									return_token_type_ids,
									return_attention_mask,
									return_overflowing_tokens,
									return_special_tokens_mask,
									return_offsets_mapping,
									return_length,
									verbose,
									cls,
									** kwargs
								)
		else:
			text_list = self.split_texts(texts, batch_size)
			results = Parallel(n_jobs=n_jobs)(
				delayed(self._order_encode)(i,
											text,
											add_special_tokens,
											padding,
											truncation,
											max_length,
											stride,
											is_split_into_words,
											pad_to_multiple_of,
											return_tensors,
											return_token_type_ids,
											return_attention_mask,
											return_overflowing_tokens,
											return_special_tokens_mask,
											return_offsets_mapping,
											return_length,
											verbose,
											cls,
											**kwargs
										) for i, text in enumerate(text_list))
			results = sorted(results, key=lambda x: x[0], reverse=False)
			results = [r[1] for r in results]
			return torch.concat(results, dim=0)

	@staticmethod
	def split_texts(texts: List[str], batch_size: int):
		return [texts[i:i + batch_size] for i in range(0, len(texts), batch_size)]

	def _order_encode(self,
				order: int,
				texts: List[str],
				add_special_tokens: bool = True,
				padding: Union[bool, str, PaddingStrategy] = False,
				truncation: Union[bool, str, TruncationStrategy] = None,
				max_length: Optional[int] = None,
				stride: int = 0,
				is_split_into_words: bool = False,
				pad_to_multiple_of: Optional[int] = None,
				return_tensors: Optional[Union[str, TensorType]] = None,
				return_token_type_ids: Optional[bool] = None,
				return_attention_mask: Optional[bool] = None,
				return_overflowing_tokens: bool = False,
				return_special_tokens_mask: bool = False,
				return_offsets_mapping: bool = False,
				return_length: bool = False,
				verbose: bool = True,
				cls: bool = False,
				**kwargs,
			) -> torch.FloatTensor:
		return order, self.encode_plus(
								texts,
								add_special_tokens,
								padding,
								truncation,
								max_length,
								stride,
								is_split_into_words,
								pad_to_multiple_of,
								return_tensors,
								return_token_type_ids,
								return_attention_mask,
								return_overflowing_tokens,
								return_special_tokens_mask,
								return_offsets_mapping,
								return_length,
								verbose,
								cls,
								**kwargs
							)


class BertTokenizeVec(TokenizeVec):

	def __init__(self, pretrained_path: Union[str, Path]):
		super().__init__(pretrained_path)
		bert_config = BertConfig.from_pretrained(pretrained_path)
		self.pretrained = BertModel.from_pretrained(pretrained_path, config=bert_config)


class ErnieTokenizeVec(TokenizeVec):

	def __init__(self, pretrained_path: Union[str, Path]):
		super().__init__(pretrained_path)
		self.pretrained = AutoModel.from_pretrained(pretrained_path)



import torch
from pathlib import Path
from typing import Union, Iterable
from torch.nn import Embedding

from ._utils import UTF8, read_file, read_corpus_files, batch_cut, token_counter, batch_pad_mask, cut, batch_pad, pad,\
    pad_mask, load_embedding


# -------------------------------------------------Tokenizer class-----------------------------------------------------
class BaseTokenizer:
    """
    encode返回的是 list
    """
    SAVE_SEP = '|'
    PAD, UNK, BOS, EOS, SEP = '<pad>', '<unk>', '<bos>', '<eos>', '<sep>'
    BACKUP_TOKENS = [PAD, UNK, BOS, EOS, SEP]

    def __init__(self, file: str = None, corpus: Iterable[str] = None, cut_corpus: Iterable[Iterable[str]] = None, vocab: Iterable[str] = None,
            min_frequency=0, reserved_token=[], language='cn', cut_type='word', word_frequency=False):
        """
        :param file: 语料文件，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
        :param corpus: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
        :param cut_corpus: 语料，每个元素是是分词后的一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
        :param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
        :param min_frequency: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
        :param reserved_token: 保留token, 如 '<pad>', '<unk>'等
        :param language: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param word_frequency: 是否统计词频
        """
        self.language = language
        self.cut_type = cut_type
        self.reserved_token = self.BACKUP_TOKENS + [token for token in reserved_token if token not in self.BACKUP_TOKENS]
        if vocab is not None:
            for token in [token for token in self.reserved_token if token in vocab]:
                vocab.remove(token)
            self.vocab = self.reserved_token + vocab
            self.token_to_idx = {k: i for i, k in enumerate(self.vocab)}
        else:
            if file is not None and corpus is None:
                corpus = read_corpus_files(file)
            if corpus is not None:
                cut_corpus = batch_cut(corpus, language=language, cut_type=cut_type, keep_punctuation=True)
            if cut_corpus is not None:
                counter = token_counter(cut_corpus)
                sorted_token_frequency = sorted(counter.items(), key=lambda kv: kv[1], reverse=True)
                self.vocab = self.reserved_token.copy()
                self.vocab += [token for token, freq in sorted_token_frequency if freq >= min_frequency and token not in self.vocab]
                self.token_to_idx = {k: i for i, k in enumerate(self.vocab)}
                if word_frequency:
                    self.word_frequency = [(self.token_to_idx[token], freq) for token, freq in sorted_token_frequency if
                                           token not in self.reserved_token]
            else:
                raise ValueError('参数file, corpus, vocab不能同时为None.')
        self.vocab_size = len(self.vocab)
        self.pad, self.unk, self.bos, self.eos, self.sep = [self.token_to_idx[token] for token in self.BACKUP_TOKENS]

    def encode(self, sentence: str, max_length: int = None, truncation=True, padding=True, padding_side='right',
            pad_begin_end=False, keep_punctuation=False, is_split_into_words: bool = False):
        """
        :param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了'
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param pad_begin_end:
        :param keep_punctuation: 是否保留标点符号
        :param is_split_into_words: 是否已经分词
        :return:
        """
        if isinstance(sentence, str):
            if is_split_into_words:
                tokens = self.do_encode(sentence)
            else:
                tokens = cut(sentence, self.language, self.cut_type, keep_punctuation=keep_punctuation)
                tokens = self.do_encode(tokens)
            return self.padding(tokens, max_length, truncation, padding, padding_side, pad_begin_end)

        raise ValueError('参数"sentence"类型错误')

    def batch_encode(self, sentences: Union[Iterable[str], Iterable[Iterable]], max_length: int = None, truncation=True, padding=True, padding_side='right',
            pad_begin_end=False, keep_punctuation=False, is_split_into_words: bool = False):
        """
        :param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param pad_begin_end:
        :param keep_punctuation: 是否保留标点符号
        :param is_split_into_words: 是否已经分词
        :return:
        """
        if isinstance(sentences, Iterable):
            if is_split_into_words:
                tokens = map(self.do_encode, sentences)
            else:
                batch_cuts = batch_cut(sentences, language=self.language, cut_type=self.cut_type, keep_punctuation=keep_punctuation)
                tokens = map(self.do_encode, batch_cuts)
            return self.padding(tokens, max_length, truncation, padding, padding_side, pad_begin_end)

        raise ValueError('参数"sentence"类型错误')

    def do_encode(self, cut_tokens: Union[str, Iterable[str]]):
        """
        把词转换成数字
        :param cut_tokens: '学生' 或 ['学生', '手机', '老师']
        :return:
        """
        if isinstance(cut_tokens, str):
            return self.token_to_idx.get(cut_tokens, self.unk)
        return list(map(self.do_encode, cut_tokens))

    def decode(self, tokens: Iterable[int], return_special_tokens=False, return_sentence=False):
        """
        :param tokens: [2, 19, 27, 3, 0, 0]
        :param return_special_tokens: 是否返回'<pad>', '<unk>', '<bos>', '<eos>'等特殊字符
        :param return_sentence: 返回的是一句话还是词序列
        :return: 由return_sentence决定，返回的是 '上课时学生手机响个不停‘, 还是 ['上课', '时', '学生', '手机', ’响个, '不停']
        """
        return [self.decode(index, return_special_tokens, return_sentence) for index in tokens]

    def batch_decode(self, tokens: Iterable[Iterable[int]], return_special_tokens=False, return_sentence=False):
        """
        :param tokens: [[2, 19, 27, 3, 0, 0], [2, 10, 3, 0, 0, 0]]
        :param return_special_tokens: 是否返回'<pad>', '<unk>', '<bos>', '<eos>'等特殊字符
        :param return_sentence: 返回的是一句话还是词序列
        :return: 由return_sentence决定，返回的是 '上课时学生手机响个不停‘, 还是 ['上课', '时', '学生', '手机', ’响个, '不停']
        """
        return [self.decode(index, return_special_tokens, return_sentence) for index in tokens]

    def padding(self, tokens: Union[map, Iterable[int], Iterable[Iterable[int]]], max_length: int, truncation=True, padding=True,
            padding_side='right', pad_begin_end=False):
        """
        :param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param pad_begin_end:
        :return:
        """
        if isinstance(tokens, map) or isinstance(tokens[0], Iterable):
            return batch_pad(tokens, max_length, truncation, padding, padding_side, pad_begin_end, self.pad, self.bos, self.eos)
        return pad(tokens, max_length, truncation, padding, padding_side, pad_begin_end, self.pad, self.bos, self.eos)

    def get_real_vocab(self):
        """
        :return: 除去特殊字符的词表
        """
        return self.vocab[len(self.reserved_token):]

    def save(self, path='vocab.txt', encoding=UTF8):
        with open(path, 'w', encoding=encoding) as f:
            f.write(self.SAVE_SEP.join([self.language, self.cut_type]) + '\n')
            f.write(self.SAVE_SEP.join(self.reserved_token) + '\n')
            f.write(self.SAVE_SEP.join(self.vocab))

    @classmethod
    def load(cls, path='vocab.txt', encoding=UTF8):
        with open(path, encoding=encoding) as f:
            lines = f.readlines()
        language, cut_type = lines[0].strip().split(Tokenizer.SAVE_SEP)
        return cls(vocab=lines[2].strip().split(Tokenizer.SAVE_SEP), reserved_token=lines[1].strip().split(Tokenizer.SAVE_SEP),
                        language=language, cut_type=cut_type)

    @classmethod
    def from_file(cls, file: str, encoding=UTF8, pattern='*', func=read_file, min_frequency=0, reserved_token=[], language='cn', cut_type='word',
            word_frequency=False):
        """
        :param file: 语料文件，可以是文件也可以是文件夹， 如：'./train.txt'. 如果是文件夹，会读取文件夹下的所有目录和文件.
        :param encoding: 编码
        :param pattern: 文件后缀，当file是文件夹的时候，会根据此后缀过滤文件
        :param func: 具体读取文件的处理函数，默认是read_file，可替换。注意：其函数签名为 function_name(path: str, encoding: str) -> corpus: Iterable[str]
        :param min_frequency: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
        :param reserved_token: 保留token, 如 '<pad>', '<unk>'等
        :param language: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param word_frequency: 是否统计词频
        """
        corpus = read_corpus_files(file, encoding, pattern, func)
        return cls(corpus=corpus, min_frequency=min_frequency, reserved_token=reserved_token, language=language, cut_type=cut_type,
                        word_frequency=word_frequency)

    @classmethod
    def from_corpus(cls, corpus: Iterable[str], min_frequency=0, reserved_token=[], language='cn', cut_type='word', word_frequency=False):
        """
        :param corpus: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param min_frequency: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
        :param reserved_token: 保留token, 如 '<pad>', '<unk>'等
        :param language: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param word_frequency: 是否统计词频
        """
        return cls(corpus=corpus, min_frequency=min_frequency, reserved_token=reserved_token, language=language, cut_type=cut_type,
                        word_frequency=word_frequency)

    @classmethod
    def from_cut_corpus(cls, cut_corpus: Iterable[Iterable[str]], min_frequency=0, reserved_token=[], language='cn', cut_type='word', word_frequency=False):
        """
        :param cut_corpus: 分词后的语料，每个元素是一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
        :param min_frequency: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
        :param reserved_token: 保留token, 如 '<pad>', '<unk>'等
        :param language: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param word_frequency: 是否统计词频
        """
        return cls(cut_corpus=cut_corpus, min_frequency=min_frequency, reserved_token=reserved_token, language=language, cut_type=cut_type,
                        word_frequency=word_frequency)

    @classmethod
    def from_vocab(cls, vocab: Iterable[str], reserved_token=[]):
        """
        :param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
        :param reserved_token: 保留token, 如 '<pad>', '<unk>'等
        """
        return cls(vocab=vocab, reserved_token=reserved_token)

    def _get_token(self, indices: Iterable[int], return_special_tokens):
        if return_special_tokens:
            return [self.vocab[i] for i in indices]
        return [self.vocab[i] for i in indices if i not in [self.pad, self.bos, self.eos, self.unk]]

    def __len__(self):
        return self.vocab_size

    def __getitem__(self, index):
        return self.vocab[index]

    # def __call__(self, sentence: Union[str, Iterable[str]], max_length: int = None, truncation=True, padding=True, padding_side='right',
    #         pad_begin_end=False, keep_punctuation=False):
    #     return self.encode(sentence, max_length, truncation, padding, padding_side, pad_begin_end, keep_punctuation)


class Tokenizer(BaseTokenizer):
    """
    encode返回的是 {'input_ids': list} 或 {'input_ids': list, 'mask_ids': list}
    """

    def __init__(self, file: str = None, corpus: Iterable[str] = None, cut_corpus: Iterable[Iterable[str]] = None, vocab: Iterable[str] = None,
            min_frequency=0, reserved_token=[], language='cn', cut_type='word', word_frequency=False):
        """
        :param file: 语料文件， 如：'./train.csv'
        :param corpus: 语料，每个元素是一句话，如：['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ...]
        :param cut_corpus: 语料，每个元素是是分词后的一句话，如：[['上课', '时', '学生', '手机', '响', '个', '不停'], ['家长', '拿', '发票', '让', '老师', '赔']]
        :param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
        :param min_frequency: 最小词频，小于词词频的词会被忽略，默认是0，所有的词都保留
        :param reserved_token: 保留token, 如 '<pad>', '<unk>'等
        :param language: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param word_frequency: 是否统计词频
        """
        super().__init__(file=file, corpus=corpus, cut_corpus=cut_corpus, vocab=vocab, min_frequency=min_frequency, reserved_token=reserved_token,
                         language=language, cut_type=cut_type, word_frequency=word_frequency)

    def encode_plus(self, sentence: Union[str, Iterable[str], Iterable[Iterable]], max_length: int = None, truncation=True, padding=True, padding_side='right',
            pad_begin_end=False, keep_punctuation=False, return_mask=False, is_split_into_words=False):
        """
        :param sentence: '上课时学生手机响个不停，老师一怒之下把手机摔了' 或 ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param pad_begin_end:
        :param keep_punctuation: 是否保留标点符号
        :param return_mask: 是否返回 mask_ids
        :param is_split_into_words: 是否已经分词
        :return:
        """
        if isinstance(sentence, str):
            if is_split_into_words:
                tokens = self.do_encode(sentence)
            else:
                tokens = cut(sentence, self.language, self.cut_type, keep_punctuation=keep_punctuation)
                tokens = self.do_encode(tokens)
            return self.padding_plus(tokens, max_length, truncation, padding, padding_side, pad_begin_end, return_mask)

        raise ValueError('参数"sentence"类型错误')

    def batch_encode_plus(self, sentences: Union[str, Iterable[str], Iterable[Iterable]], max_length: int = None, truncation=True, padding=True, padding_side='right',
            pad_begin_end=False, keep_punctuation=False, return_mask=False, is_split_into_words=False):
        """
        :param sentences: ['上课时学生手机响个不停，老师一怒之下把手机摔了。', '家长拿发票让老师赔', ....]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param pad_begin_end:
        :param keep_punctuation: 是否保留标点符号
        :param return_mask: 是否返回 mask_ids
        :param is_split_into_words: 是否已经分词
        :return:
        """
        if isinstance(sentences, Iterable):
            if is_split_into_words:
                tokens = map(self.do_encode, sentences)
            else:
                batch_cuts = batch_cut(sentences, language=self.language, cut_type=self.cut_type, keep_punctuation=keep_punctuation)
                tokens = map(self.do_encode, batch_cuts)
            return self.padding_plus(tokens, max_length, truncation, padding, padding_side, pad_begin_end, return_mask)

        raise ValueError('参数"sentence"类型错误')


    def padding_plus(self, tokens: Union[map, Iterable[int], Iterable[Iterable[int]]], max_length: int, truncation=True, padding=True,
            padding_side='right', pad_begin_end=False, return_mask=False):
        """
        :param tokens: [2, 19, 27, 3] 或 [[2, 19, 27, 3], [2, 10, 3]]
        :param max_length:
        :param truncation:
        :param padding:
        :param padding_side:
        :param pad_begin_end:
        :return:
        """
        if isinstance(tokens, map) or isinstance(tokens[0], Iterable):
            if return_mask:
                return batch_pad_mask(tokens, max_length, truncation, padding, padding_side, pad_begin_end, self.pad, self.bos, self.eos)
            return {'input_ids': batch_pad(tokens, max_length, truncation, padding, padding_side, pad_begin_end, self.pad, self.bos, self.eos)}

        if return_mask:
            return pad_mask(tokens, max_length, truncation, padding, padding_side, pad_begin_end, self.pad, self.bos, self.eos)
        return {'input_ids': pad(tokens, max_length, truncation, padding, padding_side, pad_begin_end, self.pad, self.bos, self.eos)}

    def __call__(self, sentence: Union[str, Iterable[str]], max_length: int = None, truncation=True, padding=True, padding_side='right',
            pad_begin_end=False, keep_punctuation=False, return_mask=False):
        return self.encode(sentence, max_length, truncation, padding, padding_side, pad_begin_end, keep_punctuation, return_mask)


class TokenEmbedding(Tokenizer):
    """
    可以传入已经训练好的embedding文件路径，也可以embedding数据, encode返回的是 {'input_ids': list} 或 {'input_ids': list, 'mask_ids': list}
    """

    def __init__(self, file: str = None, vocab: Iterable[str] = None, embedding: Iterable[Iterable[float]] = None, reserved_token=[],
            language='cn', cut_type='word', func=load_embedding, is_large_file=False):
        """
        :param file: embedding文件， 如：'./sgns.weibo.word.bz2'
        :param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']，与embedding必须同时传入
        :param embedding: [[0.2548, -0.6879, 0.2578],[0.2548, -0.6879, 0.2578],[0.2548, -0.6879, 0.2578]]与vocab必须同时传入
        :param reserved_token: 保留token, 如 '<pad>', '<unk>'等
        :param language: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        :param func: 具体读取文件的处理函数，load_embedding，可替换。
               注意：其函数签名为 function_name(path: str, is_large_file: bool) -> (vocab: list[str], embedding: list[list[float]])
        :param is_large_file: 是否是大文件
        """
        if file:
            assert Path(file).is_file(), 'file必须是具体文件,不能是文件夹'
            vocab, embedding = func(file, is_large_file)
        elif not vocab or not embedding:
            raise ValueError('参数"path"为空的情况下，"vocab"和"embedding"不能为空.')
        super().__init__(vocab=vocab, reserved_token=reserved_token, language=language, cut_type=cut_type)
        reserved_token = self.reserved_token.copy()
        reserved_token.reverse()
        self.dim = len(embedding[0])
        for token in reserved_token:
            embedding = [[self.do_encode(token)] * self.dim] + embedding
        self.embedding = Embedding.from_pretrained(torch.tensor(embedding, dtype=torch.float))

    def __call__(self, sentence: Union[str, Iterable[str], Iterable[Iterable]], max_length: int = None):
        input_ids = self.batch_encode(sentence, max_length)
        return self.encode(torch.tensor(input_ids, dtype=torch.long))

    @classmethod
    def from_file(cls, file: str, func=load_embedding, is_large_file=False, reserved_token=[], language='cn', cut_type='word'):
        """
        :param file: embedding文件， 如：'./sgns.weibo.word.bz2'. 注意：必须是单一文件，不能是文件夹。
        :param func: 具体读取文件的处理函数，load_embedding，可替换。注意：其函数签名为 function_name(path: str, is_large_file: bool) -> [vocab], [[embedding]]
        :param is_large_file: 是否是大文件
        :param reserved_token: 保留token, 如 '<pad>', '<unk>'等
        :param language: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        """
        return cls(file=file, reserved_token=reserved_token, language=language, cut_type=cut_type, func=func, is_large_file=is_large_file)

    @classmethod
    def from_vocab_embedding(cls, vocab: Iterable[str], embedding: Iterable[Iterable[float]], large_file=False, reserved_token=[], language='cn',
            cut_type='word'):
        """
        :param vocab: 词表，如：['上课', '学生', '手机', '不停', '，', '老师']
        :param embedding: [[0.2548, -0.6879, 0.2578],[0.2548, -0.6879, 0.2578],[0.2548, -0.6879, 0.2578]]
        :param large_file: 是否是大文件
        :param reserved_token: 保留token, 如 '<pad>', '<unk>'等
        :param language: 语言 'cn'和'en'
        :param cut_type: 分词类型，只支持'word'和‘char'两种类型
        """
        return cls(vocab=vocab, embedding=embedding, large_file=large_file, reserved_token=reserved_token, language=language, cut_type=cut_type)


if __name__ == '__main__':
    # file_name = './article.txt'
    # text = read_file(file_name)

    import pandas as pd
    import matplotlib.pyplot as plt

    DATA_PATH = r'D:\Study\kkb\代码实战课\week1\toutiao-text-classify\dataset\train.csv'
    # df = pd.read_csv(DATA_PATH)  # ['label']  ['sentence']

    # show_label_category_count(df['label'])
    # show_sentence_len_hist(df['sentence'], language='cn', cut_type='word', bins=50, scope=(0, 30))
    # show_token_frequency_plot(df['sentence'], language='cn', cut_type='word', scope=(0, 8000))

    # tokenizer = Tokenizer(corpus=df['sentence'].values, min_frequency=10, language='cn', cut_type='word')
    # tokenizer = BaseTokenizer.from_corpus(df['sentence'].values)
    # tokenizer.save()

    # sent = ('上课时学生手机响个不停，老师一怒之下把手机摔了，家长拿发票让老师赔，大家怎么看待这种事？', '老师一怒之下把手机摔了')
    # tokenizer = BaseTokenizer.load()
    # d = tokenizer.encode(sent, max_length=30, keep_punctuation=True, pad_begin_end=True)
    # print(d)
    # print(tokenizer.decode(d, return_special_tokens=False, return_sentence=True))
    # tokenizer = Tokenizer.load()
    # # print(type(tokenizer))
    # # # print(len(tokenizer))
    # # # print(tokenizer.vocab[:20])
    # # # print(tokenizer.reserved_token)
    # # # print(tokenizer.encode(['the', 'my']))
    # #
    # print(tokenizer.vocab[:20])
    # print(sent)
    # e = tokenizer(sent, max_length=30, keep_punctuation=True, pad_begin_end=True)
    # print(e)
    # d = tokenizer.decode(e)
    # print(d)

    # print(tokenizer.decode(d['input_ids'], return_special_tokens=False, return_sentence=True))
    # print(tokenizer.decode(d['input_ids']))

    # tk = [[101, 2198, 5125, 3198, 7313, 4764, 3221, 4507, 102, 7313, 4764, 3221, 4507, 102, 4764, 3221],
    #       [101, 2198, 5125, 3198, 7313, 4764, 3221, 4507, 102]]
    # r = tokenizer.padding(tk, max_length=15, truncation=True, padding_side='right', pad_begin_end=True)
    # print(r)

    # PATH = r'D:\Study\kkb\代码实战课\week1\toutiao-text-classify\dataset\sgns.weibo.word.bz2'
    # tokenizer = TokenEmbedding(PATH)
    # tokenizer = TokenEmbedding.from_file(PATH, is_large_file=False)
    # d = tokenizer.encode(sent, max_length=30, keep_punctuation=True, pad_begin_end=True)
    # print(d)
    # print(tokenizer.decode(d['input_ids'], return_special_tokens=False, return_sentence=True))
    # print(tokenizer.decode(d['input_ids']))
    #
    # tk = [[101, 2198, 5125, 3198, 7313, 4764, 3221, 4507, 102, 7313, 4764, 3221, 4507, 102, 4764, 3221],
    #       [101, 2198, 5125, 3198, 7313, 4764, 3221, 4507, 102]]
    # r = tokenizer.padding(tk, max_length=15, truncation=True, padding_side='right', pad_begin_end=True)
    # print(r)

    # print(batch_pad_mask(tk))

    # tokenizer = BaseTokenizer.from_file('./', pattern='*.txt', language='en')
    # print(tokenizer.vocab)

    # for f in get_files(r'D:\tmp'):
    #     print(f)

    # for line in read_large_file('./article.txt'):
    #     print(line)




import torch
import numpy as np
from torch import optim
from pathlib import Path
from typing import Union, List
from torch.nn import functional as F
from torch.utils.data import DataLoader, TensorDataset, Dataset

from nlpx.text_token import get_texts_max_length
from nlpx.training import Trainer, SimpleTrainer, evaluate


class ModelWrapper:

	def __init__(self, model_path: Union[str, Path], classes: List[str] = None,
					device=torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')):
		self.classes = classes
		self.device = device
		self.model = torch.load(model_path, map_location=device) if model_path else None

	def train(self, train_set: Dataset, eval_set: Dataset, collate_fn=None, max_iter=100,
				optimizer=optim.AdamW, learning_rate=0.001, T_max: int = 0,
				batch_size=8, eval_batch_size=16,
				num_workers=4, num_eval_workers=2,
				early_stopping_rounds=10):
		trainer = Trainer(self.model, train_set, eval_set, collate_fn,
						  max_iter, optimizer, learning_rate, T_max,
						  batch_size, eval_batch_size,
						  num_workers, num_eval_workers,
						  early_stopping_rounds,  # 早停，等10轮决策，评价指标不在变化，停止
						  self.device)
		self.model = trainer.train()

	def predict(self, X: torch.FloatTensor):
		logits = self.logits(X)
		return logits.argmax(1)

	def predict_classes(self, X: torch.FloatTensor):
		assert self.classes is not None, 'classes must be specified'
		pred = self.predict(X)
		return [self.classes[i] for i in pred.detach().numpy().ravel()]

	def predict_proba(self, X: torch.FloatTensor):
		logits = self.logits(X)
		result = F.softmax(logits).max(1)
		return result.indices, result.values

	def predict_classes_proba(self, X: torch.FloatTensor):
		assert self.classes is not None, 'classes must be specified'
		logits = self.logits(X)
		result = F.softmax(logits).max(1)
		return [self.classes[i] for i in result.indices.detach().numpy().ravel()], result.values

	def logits(self, X: torch.FloatTensor):
		self.model.eval()
		with torch.no_grad():
			logits = self.model(X)
		return logits

	def evaluate(self, eval_set: Dataset, batch_size=16, num_workers=0, max_length: int = 0,  collate_fn=None):
		eval_loader = DataLoader(dataset=eval_set, batch_size=batch_size,
								 num_workers=num_workers, collate_fn=collate_fn)
		_, acc = evaluate(self.model, eval_loader, self.device)
		return acc

	def save(self, model_path: Union[str, Path] = './best_model.pt'):
		torch.save(model_path, self.model)

	def load(self, model_path: Union[str, Path] = './best_model.pt'):
		self.model = torch.load(model_path, map_location=self.device)


class SimpleModelWrapper(ModelWrapper):

	def __init__(self, tokenize_vec, model_path: Union[str, Path] = None, classes: List[str] = None,
					device=torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')):
		super().__init__(model_path, classes, device)
		self.tokenize_vec = tokenize_vec

	def train(self, model, texts: List[str], y: Union[torch.LongTensor, List, np.ndarray], max_length: int = 0,
			  eval_size=0.2, random_state=None, collate_fn=None,max_iter=100,
			  optimizer=optim.AdamW, learning_rate=0.001, T_max: int = 0,
			  batch_size=8, eval_batch_size=16,
			  num_workers=4, num_eval_workers=2,
			  early_stopping_rounds=10):
		max_length = self.get_max_length(texts, max_length)
		X = self.get_vec(texts, max_length=max_length)
		if isinstance(y, List) or isinstance(y, np.ndarray):
			y = torch.tensor(y, dtype=torch.long)
		trainer = SimpleTrainer(model, X, y, eval_size, random_state, collate_fn,
								 max_iter, optimizer, learning_rate, T_max,
								 batch_size, eval_batch_size,
								 num_workers, num_eval_workers,
								 early_stopping_rounds,
								 self.device)
		self.model = trainer.train()

	def predict(self, texts: List[str], max_length: int = 0):
		logits = self.logits(texts, max_length)
		return logits.argmax(1)

	def predict_classes(self, texts: List[str], max_length: int = 0):
		assert self.classes is not None, 'classes must be specified'
		pred = self.predict(texts, max_length)
		return [self.classes[i] for i in pred.detach().numpy().ravel()]
	
	def predict_proba(self, texts: List[str], max_length: int = 0):
		logits = self.logits(texts, max_length)
		result = F.softmax(logits).max(1)
		return result.indices, result.values

	def predict_classes_proba(self, texts: List[str], max_length: int = 0):
		assert self.classes is not None, 'classes must be specified'
		logits = self.logits(texts, max_length)
		result = F.softmax(logits).max(1)
		return [self.classes[i] for i in result.indices.detach().numpy().ravel()], result.values

	def logits(self, texts: List[str], max_length: int = 0):
		X = self.get_features(texts, max_length)
		return super().logits(X)

	def evaluate(self, texts: List[str], y: Union[torch.LongTensor, List, np.ndarray], batch_size=16, num_workers=0,
				 max_length: int = 0, collate_fn=None):
		X = self.get_features(texts, max_length)
		if isinstance(y, List) or isinstance(y, np.ndarray):
			y = torch.tensor(y, dtype=torch.long)
		eval_set = TensorDataset(X, y)
		return super().evaluate(eval_set, batch_size=batch_size, num_workers=num_workers, collate_fn=collate_fn)

	@staticmethod
	def get_max_length(texts: List[str], max_length: int = 0) -> int:
		if max_length and max_length > 0:
			return max_length
		return get_texts_max_length(texts, cut_type='char')

	def get_vec(self, texts: List[str], max_length: int):
		return self.tokenize_vec.parallel_encode_plus(texts, max_length=max_length, padding='max_length',
														truncation=True, add_special_tokens=True,
														return_token_type_ids=True,return_attention_mask=True,
														return_tensors='pt')

	def get_features(self, texts: List[str], max_length: int = 0):
		max_length = self.get_max_length(texts, max_length)
		return self.get_vec(texts, max_length=max_length)



import torch
from torch import nn
from torch.nn import functional as F


class TextCNN(nn.Module):

    def __init__(self, word_dim: int, kernel_sizes=(2, 3, 4), cnn_channels: int = 64, out_features: int = 2,
                 activation=nn.ReLU(inplace=True), num_hidden_layer: int = 0, drop_out: float = 0.0):
        """ TextCNN model
        Parameters
        ----------
        word_dim: int, dim of word, in_channels of cnn
        cnn_channels: int, out_channels of cnn
        kernel_sizes: size of each cnn kernel
        out_features: dim of output
        activation:
        num_hidden_layer:
        drop_out:

        Examples
        --------
        >>> import torch
        >>> from nlpx.models import TextCNN
        >>> X = torch.randn(batch_size, 10, word_dim)
        >>> targets = torch.randint(0, num_classes, (batch_size,))
        >>> model = TextCNN(word_dim, cnn_channels=64, kernel_sizes=(2, 3, 4), out_features=num_classes)
        >>> output = model(X)
        >>> loss, output = model(X, targets)
        """
        super().__init__()
        self.convs = nn.ModuleList([
            nn.Sequential(
                nn.Conv1d(in_channels=word_dim, out_channels=cnn_channels, kernel_size=kernel_size, bias=False),
                nn.BatchNorm1d(num_features=cnn_channels),
                activation,  # inplace为True，将会改变输入的数据 ，否则不会改变原输入，只会产生新的输出
                nn.AdaptiveMaxPool1d(1)
            ) for kernel_size in kernel_sizes
        ])

        self.num_hidden_layer = num_hidden_layer
        num_features = cnn_channels * len(kernel_sizes)
        if num_hidden_layer and num_hidden_layer > 0:
            self.hidden_layers = nn.ModuleList([
                nn.Sequential(
                    nn.Linear(in_features=num_features, out_features=num_features),
                    nn.LayerNorm(normalized_shape=num_features),
                    activation
                ) for _ in range(num_hidden_layer)
            ])

        self.fc = nn.Linear(in_features=num_features, out_features=out_features)
        if 0.0 < drop_out < 1.0:
            self.fc = nn.Sequential(
                nn.Dropout(drop_out),
                self.fc
            )

    def forward(self, inputs, labels=None):
        """
        :param inputs: [(batch, sentence, word_dim)]
        :param labels: [long]
        """
        input_embeddings = inputs.transpose(2, 1)
        out = torch.cat([conv(input_embeddings) for conv in self.convs], dim=1)
        out = out.transpose(2, 1)

        if self.num_hidden_layer and self.num_hidden_layer > 0:
            for hidden_layer in self.hidden_layers:
                out = hidden_layer(out)

        out = self.fc(out)
        logits = out.squeeze(1)

        if labels is None:
            return logits
        else:
            loss = F.cross_entropy(logits, labels)
            return loss, logits



import torch
from torch import optim
from torch.utils.data import DataLoader, TensorDataset, Dataset
from torch.optim.lr_scheduler import CosineAnnealingLR
from sklearn.model_selection import train_test_split


def evaluate(model, eval_loader, device):
	total = 0.0
	losses = 0.0
	correct = 0.0
	model.eval()
	with torch.no_grad():
		for x, y in eval_loader:
			x, y = x.to(device), y.to(device)
			loss, logits = model(x, y)
			correct += (logits.argmax(1) == y).sum().item()
			losses += loss.item()
			total += len(y)
	return losses / total, correct / total


class Trainer:

	def __init__(self, model, train_set: Dataset, eval_set: Dataset, collate_fn=None,
				 max_iter=100, optimizer=optim.AdamW, learning_rate=0.001, T_max: int = 0,
				 batch_size=8, eval_batch_size=16,
				 num_workers=4, num_eval_workers=2,
				 early_stopping_rounds=10,  # 早停，等10轮决策，评价指标不在变化，停止
				 device=torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
				 ):
		self.model = model.to(device)
		self.train_loader = DataLoader(dataset=train_set, batch_size=batch_size,
									   num_workers=num_workers, shuffle=True, collate_fn=collate_fn)
		self.eval_loader = DataLoader(dataset=eval_set, batch_size=eval_batch_size,
									  num_workers=num_eval_workers, collate_fn=collate_fn)
		self.max_iter = max_iter
		self.optimizer = optimizer(model.parameters(), lr=learning_rate)
		self.T_max = T_max
		self.early_stopping_rounds = early_stopping_rounds
		self.device = device

	def train(self):
		cnt = 0
		cnt2 = 0
		best_acc = 0.0
		last_acc = 0.0
		best_model = None
		if self.T_max and self.T_max > 0:
			scheduler = CosineAnnealingLR(self.optimizer, T_max=self.T_max)
		for epoch in range(self.max_iter):
			total = 0.0
			losses = 0.0
			correct = 0.0
			self.model.train()
			for x, y in self.train_loader:
				x, y = x.to(self.device), y.to(self.device)
				loss, logits = self.model(x, y)
				self.optimizer.zero_grad()
				loss.backward()
				self.optimizer.step()
				if self.T_max and self.T_max > 0:
					scheduler.step()

				losses += loss.item()
				correct += (logits.argmax(1) == y).sum().item()
				total += len(y)

			val_loss, val_acc = evaluate(self.model, self.eval_loader, self.device)
			print('epoch-{}/{}\tlr: {:.6f}, train_loss: {:.4f}, train_acc: {:.4f}, val_loss: {:.4f}, val_acc: {:.4f}'.format(
				epoch + 1, self.max_iter, self.optimizer.param_groups[0]['lr'], losses / total, correct / total, val_loss, val_acc
			))
			if epoch > 5 and val_acc > best_acc:
				best_acc = val_acc
				best_model = self.model

			if val_acc >= best_acc:  # val_acc提升
				cnt = 0

			if last_acc == val_acc:  # val_acc不在变化
				cnt2 += 1
			else:
				cnt2 = 0

			last_acc = val_acc
			cnt += 1
			if epoch >= min(5, self.early_stopping_rounds) and max(cnt, cnt2) > self.early_stopping_rounds:  # x次epoch的val_acc不提升或x次epoch的val_acc不变化
				print("Early stopping at epoch-{}/{}".format(epoch + 1, self.max_iter))
				break

		return best_model


class SimpleTrainer(Trainer):

	def __init__(self, model, X, y, eval_size=0.2, random_state=None, collate_fn=None,
				 max_iter=100, optimizer=optim.AdamW, learning_rate=0.001, T_max: int = 0,
				 batch_size=8, eval_batch_size=16,
				 num_workers=4, num_eval_workers=2,
				 early_stopping_rounds=10,  # 早停，等10轮决策，评价指标不在变化，停止
				 device=torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')):
		X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=eval_size, random_state=random_state)
		super().__init__(model,
						 TensorDataset(X_train, y_train),
						 TensorDataset(X_test, y_test),
						 collate_fn,
						 max_iter,
						 optimizer,
						 learning_rate,
						 T_max,
						 batch_size,
						 eval_batch_size,
						 num_workers,
						 num_eval_workers,
						 early_stopping_rounds,
						 device
						 )



from ._utils import read_file, read_large_file, read_corpus_files, get_texts_max_length, token_counter, \
	show_label_category_count, show_sentence_len_hist, show_token_frequency_plot, cut, batch_cut, pad, batch_pad, \
	batch_pad_mask, load_embedding
from ._tokenizer import BaseTokenizer, Tokenizer, TokenEmbedding
